# -*- coding: utf-8 -*-
import os
import json
from mpcontribs.io.core import mp_level01_titles
from mpcontribs.io.core.recdict import RecursiveDict
from mpcontribs.io.core.components.hdata import HierarchicalData
from mpcontribs.io.core.components.tdata import Table
from mpcontribs.io.core.components.gdata import Plot


def test_components():
    d = {k: {} for k in mp_level01_titles}
    d["data"] = {"a": 3.5, "b": {"display": "6 eV", "value": 6, "unit": "eV"}}
    exp = HierarchicalData([("data", RecursiveDict([("a", 3.5), ("b", "6 eV")]))])
    assert HierarchicalData(d) == exp

    fn = os.path.join(os.path.dirname(__file__), "test_table.json")
    with open(fn, "r") as f:
        d = json.load(f)
        table = Table.from_dict(d)
        assert table.render()
        plot = Plot.from_dict(d)
        assert plot.get_figure()



# -*- coding: utf-8 -*-
import re

mp_id_pattern = re.compile(r"^(mp|por|mvc)-\d+(?:--\d+)?$", re.IGNORECASE)
replacements = {" ": "_", "[": "", "]": "", "{": "", "}": "", ":": "_"}
mp_level01_titles = ["_hdata", "tables", "graphs", "structures"]



# -*- coding: utf-8 -*-
import os
from mpcontribs.io.archie.mpfile import MPFile


def test_get_string():
    test_file = os.path.join(os.path.dirname(__file__), "test_archieml.txt")
    mpfile = MPFile.from_file(test_file)
    mpfile_test = MPFile.from_string(mpfile.get_string())
    assert mpfile_test
    # assert mpfile.document == mpfile_test.document



# -*- coding: utf-8 -*-
import uuid
import json
import pandas as pd
from IPython.display import display_html
from mpcontribs.io.core import mp_id_pattern
from mpcontribs.io.core.utils import clean_value
from mpcontribs.io.core.recdict import RecursiveDict
from urllib.parse import urlparse


class Table(pd.DataFrame):
    def __init__(
        self,
        data,
        columns=None,
        index=None,
        api_key=None,
        ncols=12,
        per_page=None,
        filters=None,
        **kwargs,
    ):
        if columns is None and isinstance(data, list):
            columns = list(data[0].keys())
        super(Table, self).__init__(data=data, index=index, columns=columns)
        self.tid = kwargs.get("id")
        self.api_key = api_key
        self.project = kwargs.get("project")
        self.name = kwargs.get("name", "Table")
        self.cid = kwargs.get("contribution")
        self.ncols = ncols
        self.per_page = per_page
        self.filters = filters

    def to_dict(self):
        from pandas import MultiIndex

        df = self.copy()
        if not df.index.empty and not isinstance(df.index, MultiIndex):
            df = self.reset_index()
        for col in df.columns:
            df[col] = df[col].apply(lambda x: clean_value(x, max_dgts=6))
        rdct = df.to_dict(orient="split", into=RecursiveDict)
        if not isinstance(df.index, MultiIndex):
            rdct.pop("index")
        rdct["contribution"] = self.cid
        rdct["name"] = self.name
        return rdct

    @classmethod
    def from_dict(cls, d):
        index = None
        if "index" in d:
            from pandas import MultiIndex

            index = MultiIndex.from_tuples(d["index"])
        skip = {"data", "columns", "index"}
        kwargs = dict((k, v) for k, v in d.items() if k not in skip)
        return cls(d["data"], columns=d["columns"], index=index, **kwargs)

    @classmethod
    def from_items(cls, rdct, **kwargs):
        return super(Table, cls).from_dict(RecursiveDict(rdct), **kwargs)

    def to_backgrid_dict(self):
        """Backgrid-conform dict from DataFrame"""
        # shorten global import times by importing django here
        from mpcontribs.io.core.utils import get_composition_from_string
        from pandas import MultiIndex
        import pymatgen.util as pmg_util
        from pymatgen.core.composition import CompositionError

        table = dict()
        nrows_max = 260
        nrows = self.shape[0]
        df = Table(self.head(n=nrows_max)) if nrows > nrows_max else self

        if isinstance(df.index, MultiIndex):
            df.reset_index(inplace=True)

        table["columns"] = []
        table["rows"] = super(Table, df).to_dict(orient="records")

        for col_index, col in enumerate(list(df.columns)):
            cell_type = "number"

            # avoid looping rows to minimize use of `df.iat` (time-consuming in 3d)
            if not col.startswith("level_") and col[-1] != "]":
                is_url_column = True

                for row_index in range(df.shape[0]):
                    cell = str(df.iat[row_index, col_index])
                    is_url_column = bool(
                        is_url_column and (not cell or mp_id_pattern.match(cell))
                    )
                    if is_url_column:
                        if cell:
                            value = "https://materialsproject.org/materials/{}".format(
                                cell
                            )
                            table["rows"][row_index][col] = value
                    elif cell:
                        try:
                            composition = get_composition_from_string(cell)
                            composition = pmg_util.string.unicodeify(composition)
                            table["rows"][row_index][col] = composition
                        except (CompositionError, ValueError, OverflowError):
                            try:
                                # https://stackoverflow.com/a/38020041
                                result = urlparse(cell)
                                if not all([result.scheme, result.netloc, result.path]):
                                    break
                                is_url_column = True
                            except Exception:
                                break

                cell_type = "uri" if is_url_column else "string"

            col_split = col.split(".")
            nesting = [col_split[0]] if len(col_split) > 1 else []
            table["columns"].append(
                {"name": col, "cell": cell_type, "nesting": nesting, "editable": 0}
            )
            if len(col_split) > 1:
                table["columns"][-1].update({"label": ".".join(col_split[1:])})

        return table

    def render(self, total_records=None):
        """use BackGrid JS library to render Pandas DataFrame"""
        # if project given, this will result in an overview table of contributions
        # TODO check for index column in df other than the default numbering
        jtable = json.dumps(self.to_backgrid_dict())
        if total_records is None:
            total_records = self.shape[0]
        config = {"total_records": total_records}
        config["uuids"] = [str(uuid.uuid4()) for i in range(4)]
        if self.tid:
            config["tid"] = self.tid
            config["per_page"] = self.per_page
        else:
            config["project"] = self.project
        config["api_key"] = self.api_key
        config["ncols"] = self.ncols
        config["filters"] = self.filters
        jconfig = json.dumps(config)
        html = '<div class="col-md-6" id="{}"></div>'.format(config["uuids"][0])
        html += '<div class="pull-right" id="{}"></div>'.format(config["uuids"][3])
        html += '<div id="{}" style="width:100%;"></div>'.format(config["uuids"][1])
        html += '<div id="{}"></div>'.format(config["uuids"][2])
        html += f"<script>render_table({{table: {jtable}, config: {jconfig}}})</script>"
        return html

    def _ipython_display_(self):
        display_html(self.render(), raw=True)


class Tables(RecursiveDict):
    """class to hold and display multiple data tables"""

    def __init__(self, content=RecursiveDict()):
        super(Tables, self).__init__(
            (key, value) for key, value in content.items() if isinstance(value, Table)
        )

    def __str__(self):
        return "tables: {}".format(" ".join(self.keys()))

    def _ipython_display_(self):
        for name, table in self.items():
            display_html("<h3>{}</h3>".format(name), raw=True)
            display_html(table)


class TabularData(RecursiveDict):
    """class to hold and display all tabular data of a MPFile"""

    def __init__(self, document):
        super(TabularData, self).__init__()
        from pymatgen import Structure

        scope = []
        for key, value in document.iterate():
            if isinstance(value, Table):
                self[scope[0]].rec_update({".".join(scope[1:]): value})
            elif not isinstance(value, Structure):
                level, key = key
                level_reduction = bool(level < len(scope))
                if level_reduction:
                    del scope[level:]
                if value is None:
                    scope.append(key)
                    if scope[0] not in self:
                        self[scope[0]] = Tables()

    def __str__(self):
        return "mp-ids: {}".format(" ".join(self.keys()))

    def _ipython_display_(self):
        for identifier, tables in self.items():
            if isinstance(tables, dict) and tables:
                display_html(
                    "<h2>Tabular Data for {}</h2>".format(identifier), raw=True
                )
                display_html(tables)



# -*- coding: utf-8 -*-
import json
import uuid
import numpy as np
import plotly.io as pio
from plotly.io._utils import validate_coerce_fig_to_dict
from plotly.offline.offline import _get_jconfig
from plotly.utils import PlotlyJSONEncoder
from IPython.display import display_html
from mpcontribs.io.core import mp_level01_titles
from mpcontribs.io.core.recdict import RecursiveDict
from mpcontribs.io.core.components.tdata import Table


class MyRenderer(pio.base_renderers.MimetypeRenderer):
    tid = None

    def to_mimebundle(self, fig):
        fig_dict = validate_coerce_fig_to_dict(fig, True)
        divid = str(uuid.uuid4())
        data = fig_dict.get("data", [])
        jdata = json.dumps(data, cls=PlotlyJSONEncoder, sort_keys=True)
        layout = fig_dict.get("layout", {})
        jlayout = json.dumps(layout, cls=PlotlyJSONEncoder, sort_keys=True)
        config = _get_jconfig(None)
        config.setdefault("responsive", True)
        jconfig = json.dumps(config)
        script = f'render_plot({{divid: "{divid}", layout: {jlayout}, config: {jconfig}'
        script += f', tid: "{self.tid}", data: {jdata}}})'
        html = f'<div><div id="{divid}"></div>'
        html += f'<script type="text/javascript">{script};</script></div>'
        return {"text/html": html}


my_renderer = MyRenderer()
pio.renderers["my_renderer"] = my_renderer
pio.renderers.render_on_display = True


class Plot(object):
    """class to hold and display single interactive graph/plot"""

    def __init__(self, table, config=None):
        self.table = table
        self.config = config or {}

    @classmethod
    def from_dict(cls, d):
        return cls(Table.from_dict(d), config=d.get("config"))

    def get_figure(self):
        from pandas import MultiIndex

        layout = dict(legend=dict(x=0.7, y=1), margin=dict(r=0, t=40))
        is_3d = isinstance(self.table.index, MultiIndex)
        if is_3d:
            import plotly.graph_objs as go
            from plotly import tools

            cols = self.table.columns
            ncols = 2 if len(cols) > 1 else 1
            nrows = len(cols) / ncols + len(cols) % ncols
            fig = tools.make_subplots(
                rows=nrows, cols=ncols, subplot_titles=cols, print_grid=False
            )
            for idx, col in enumerate(cols):
                series = self.table[col]
                z = [s.tolist() for _, s in series.groupby(level=0)]
                fig.append_trace(
                    go.Heatmap(z=z, showscale=False), idx / ncols + 1, idx % ncols + 1
                )
            fig["layout"].update(layout)
        else:
            xaxis = self.config.get("x", self.table.columns[0])
            yaxis = self.config.get("y", None)
            yaxes = (
                [yaxis]
                if yaxis is not None
                else [col for col in self.table.columns if col != xaxis]
            )
            traces = []
            for axis in yaxes:
                if "ₑᵣᵣ" not in axis:
                    tbl = self.table[[xaxis, axis]].replace("", np.nan).dropna()
                    traces.append(
                        dict(x=tbl[xaxis].tolist(), y=tbl[axis].tolist(), name=axis)
                    )
            for trace in traces:
                err_axis = trace["name"] + "ₑᵣᵣ"
                if err_axis in yaxes:
                    errors = self.table[err_axis].replace("", np.nan).dropna()
                    trace["error_y"] = dict(type="data", array=errors, visible=True)
                    trace["mode"] = "markers"
            layout.update(
                dict(
                    xaxis=dict(title=xaxis),
                    yaxis=dict(
                        title=self.config.get("ytitle"),
                        type=self.config.get("yaxis", {}).get("type", "-"),
                    ),
                    showlegend=self.config.get("showlegend", True),
                )
            )
            fig = dict(data=traces, layout=layout)

        return fig

    def _ipython_display_(self):
        fig = self.get_figure()
        is_3d = False  # TODO decide by heatmap
        axis = "z" if is_3d else "x"
        npts = len(fig.get("data")[0][axis])
        static_fig = (is_3d and npts > 150) or (not is_3d and npts > 700)
        renderers = ["jupyterlab"]
        renderers.append("png" if static_fig else "my_renderer")
        pio.renderers.default = "+".join(renderers)
        pio.show(fig, tid=self.table.tid)


class Plots(RecursiveDict):
    """class to hold and display multiple interactive graphs/plots"""

    def __init__(self, tables, plotconfs):
        super(Plots, self).__init__(
            (plotconf["table"], Plot(plotconf, tables[plotconf["table"]]))
            for plotconf in plotconfs.values()
        )

    def __str__(self):
        return "plots: {}".format(" ".join(self.keys()))

    def _ipython_display_(self):
        for name, plot in self.items():
            if plot:
                display_html("<h3>{}</h3>".format(name), raw=True)
                display_html(plot)


class GraphicalData(RecursiveDict):
    """class to hold and display all interactive graphs/plots of a MPFile"""

    def __init__(self, document):
        from mpcontribs.io.core.components.tdata import TabularData

        tdata = TabularData(document)
        super(GraphicalData, self).__init__(
            (identifier, Plots(tdata[identifier], content[mp_level01_titles[2]]))
            for identifier, content in document.items()
            if mp_level01_titles[2] in content
        )

    def __str__(self):
        return "mp-ids: {}".format(" ".join(self.keys()))

    def _ipython_display_(self):
        for identifier, plots in self.items():
            if identifier != mp_level01_titles[0] and plots:
                display_html(
                    "<h2>Interactive Plots for {}</h2>".format(identifier), raw=True
                )
                display_html(plots)



# -*- coding: utf-8 -*-
from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="mpcontribs-io",
    author="Patrick Huck",
    author_email="phuck@lbl.gov",
    description="MPContribs I/O Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/materialsproject/MPContribs/tree/master/mpcontribs-io",
    packages=["mpcontribs.io.core", "mpcontribs.io.archie"],
    install_requires=["archieml", "ipython", "pandas", "plotly", "pymatgen"],
    license="MIT",
    zip_safe=False,
    include_package_data=True,
    use_scm_version={"root": "..", "relative_to": __file__},
    setup_requires=["setuptools_scm"],
)
